#ifndef __PARSE_H
#define __PARSE_H

#include <cstdio>
#include <cstdlib>
#include <string>
#include <algorithm>
#include <stack>

#include <stxxl/bits/containers/sorter.h>
#include <stxxl/bits/containers/stack.h>

#include "stream.h"
#include "uint40.h"
#include "radixsort.h"
#include "utils.h"

#define stxxl_block_size (1024 * 1024)

struct LpfTriple {
  LpfTriple() {}
  LpfTriple(uint40 idx, uint40 pos, uint40 len)
    : m_idx(idx), m_pos(pos), m_len(len) {}

  uint40 m_idx, m_pos, m_len;
};

struct CmpLpfTriple {
  inline bool operator() (const LpfTriple &a, const LpfTriple &b) const {
    return a.m_idx < b.m_idx;
  }

  LpfTriple min_value() const {
    return LpfTriple(std::numeric_limits<uint40>::min(), 0L, 0L);
  }

  LpfTriple max_value() const {
    return LpfTriple(std::numeric_limits<uint40>::max(), 0L, 0L);
  }
};

namespace stxxl {
namespace algorithm {

template <typename Iterator>
struct Sort<Iterator, CmpLpfTriple> {
    static inline void sort(Iterator begin, Iterator end, CmpLpfTriple& cmp) {
        radixsort8msb_copy2(begin, end, sizeof(uint40) - 1, cmp);
    }
};

} // namespace algorithm
} // namespace stxxl

inline long max_alloc() {
  return stxxl::block_manager::get_instance()->max_allocated();
}

void parse_file(std::string in_fname, long ram_use, std::string out_fname) {
  long double time_start = utils::wclock();

  long length = utils::file_size(in_fname);
  long double length_MiB = (long double)length / (1 << 20);

  fprintf(stderr, "Text filename = %s\n", in_fname.c_str());
  fprintf(stderr, "Text length = %lu\n", length);

  stream_reader<uint40> *sa = new stream_reader<uint40>(in_fname + ".sa5");
  stream_reader<uint40> *lcp = new stream_reader<uint40>(in_fname + ".lcp5");

  CmpLpfTriple cmp_lcp;
  stxxl::sorter<LpfTriple, CmpLpfTriple, stxxl_block_size>
    sorter(cmp_lcp, (20L * ram_use) / 31L);

  typedef stxxl::STACK_GENERATOR<std::pair<long, long> >::result stack_type;
  stack_type *s = new stack_type();
  s->push(std::make_pair(-1L, -1L));

  long double scanning_start = utils::wclock();
  for (long r = 0, dbg = 0; r < length; ++r, ++dbg) {
    long r_lcp = lcp->read().ll();
    long SAr = sa->read().ll();

    while (!(s->empty())) {
      long SAt = s->top().first;
      long t_lcp = s->top().second;

      if (SAr < SAt) {
        s->pop();
        if (r_lcp < t_lcp)
          sorter.push(LpfTriple(SAt, s->top().first, t_lcp));
        else sorter.push(LpfTriple(SAt, SAr, r_lcp));

        r_lcp = std::min(t_lcp, r_lcp);
      } else if (SAr > SAt && r_lcp <= t_lcp) {
        s->pop();

        sorter.push(LpfTriple(SAt, s->top().first, t_lcp));
      } else break;
    }

    s->push(std::make_pair(SAr, r_lcp));

    if (dbg == (1 << 23)) {
      dbg = 0;

      long double elapsed = utils::wclock() - scanning_start;
      long double scanned_m = r / (1024.L * 1024);
      long inp_vol = 2 * sizeof(uint40) * r;
      long out_vol = sizeof(LpfTriple) * sorter.size();
      long tot_vol = inp_vol + out_vol;
      long double tot_vol_m = tot_vol / (1024.L * 1024);
      long double io_speed = tot_vol_m / elapsed;
      fprintf(stderr, "Scanned %.1LfM SA values (%.1Lf%%), "
          "time = %.1Lfs (%.3Lfs/iMiB), io = %.2LfMiB/s\r",
          scanned_m, (100.L * r) / length, elapsed,
          elapsed / scanned_m, io_speed);
    }
  }

  while (s->top().first != -1) {
    long SAt = s->top().first;
    long t_lcp = s->top().second;

    s->pop();
    sorter.push(LpfTriple(SAt, s->top().first, t_lcp));
  }

  delete sa;
  delete lcp;
  delete s;

  long double scanning_time = utils::wclock() - scanning_start;
  fprintf(stderr, "\nScanned SA in %.1Lfs (%.3Lfs/iMiB), max-alloc = %.2Lfbytes/char\n",
      scanning_time, scanning_time / length_MiB, (1.L * max_alloc()) / length);

  long next = 0, dbg = 0, curpos = 0, n_phrase = 0, out_vol = 0;

  // Open output file.
  bool write_out = (out_fname.empty() == false);
  stream_writer<uint40> *out_writer = NULL;
  if (write_out)
    out_writer = new stream_writer<uint40>(out_fname, 4 << 20);

  // Merge runs of LPF.
  long double merging_start = utils::wclock();
  sorter.sort();
  FILE *ftxt = utils::open_file(in_fname, "r");
  while (!sorter.empty()) {
    LpfTriple triple = *sorter;

    if (curpos == next) {
      ++n_phrase;
      uint40 len = triple.m_len;
      uint40 pos = triple.m_pos;

      if (!len)
        pos = utils::read_from_pos<unsigned char>(ftxt, curpos);

      if (write_out) {
        out_writer->write(pos);
        out_writer->write(len);
        out_vol += 2L * sizeof(uint40);
      }

      next += std::max(1UL, len.ull());
    }

    ++dbg;
    ++curpos;
    ++sorter;

    // Print progress.
    if (dbg == (1 << 23)) {
      dbg = 0;

      long double elapsed = utils::wclock() - merging_start;
      long double scanned_m = curpos / (1024.L * 1024);
      long inp_vol = sizeof(LpfTriple) * curpos;
      long tot_vol = inp_vol + out_vol;
      long double tot_vol_m = tot_vol / (1024.L * 1024);
      long double io_speed = tot_vol_m / elapsed;
      fprintf(stderr, "Merging: processed %.1LfM elements (%.1Lf%%), "
          "time = %.1Lfs (%.3Lfs/iMiB), io = %.2LfMiB/s\r",
          scanned_m, (100.L * curpos) / length, elapsed,
          elapsed / scanned_m, io_speed);
    }
  }

  sorter.finish_clear();
  std::fclose(ftxt);

  if (out_writer)
    delete out_writer;

  long double merging_time = utils::wclock() - merging_start;
  fprintf(stderr, "\nMerging: time = %.1Lfs (%.3Lfs/iMiB), max-alloc = %.2Lfbytes/char\n",
      merging_time, merging_time / length_MiB, 1.L * max_alloc() / length);

  long double total_time = utils::wclock() - time_start;
  fprintf(stderr, "Total time: %.2Lf (%.3Lfs/iMiB)\n",
      total_time, total_time / length_MiB);

  fprintf(stderr, "Number of phrases = %ld\n", n_phrase);
}

#endif // __PARSE_H
